(define (basic-block-set->list bb-set entry-lbl)
  (let ((bb-list (queue->list (vector-ref bb-set 1)))
        (bb-queue (queue-empty)))

    (define (find-block lbl)
      (let loop ((l bb-list))
        (if (= (cadr (basic-block-label (car l))) lbl)
          (car l)
          (loop (cdr l)))))

    (define (remove-bb bb bb-left)
      (if (eq? (car bb-left) bb)
        (cdr bb-left)
        (cons (car bb-left) (remove-bb bb (cdr bb-left)))))

    ; return a basic block which ends with a branch to 'bb' (and that is
    ; still in 'bb-left') or #f if there aren't any
    (define (prec-bb bb bb-left)
      (let loop ((l (basic-block-precedents bb)))
        (if (null? l)
          #f
          (let ((prec (find-block (car l))))
            (if (memq prec bb-left) prec (loop (cdr l)))))))

    ; return the basic block which 'bb' jumps to (and that is still in
    ; 'bb-left') or #f if there aren't any
    (define (next-bb bb bb-left)
      (let* ((branch (basic-block-branch bb))
             (branch-name (car branch))
             (next (if (eq? branch-name 'JUMP)
                     (let ((dest (cadr branch)))
                       (and (lbl? dest) (find-block lbl-num dest))))
                     (let ((block (find-block (basic-block-fall-through bb))))
                       (if (memq block bb-left)
                         block
                         (find-block (car (cddddr branch))))))))
        (and next (memq next bb-left) next))

    (define (insert-now! bb bb-left)
      (queue-put! bb-queue bb)
      (insert-after! bb (remove-bb bb bb-left)))

    (define (insert-after! bb bb-left)
      (if (not (null? bb-left))
        (if bb
          (let ((next (next-bb bb bb-left)))
            (insert-later! next bb-left insert-after!))
          (insert-now! (car bb-left) bb-left))))

    (define (insert-later! bb bb-left insert!)
      (if bb
        (let ((bb-left (remove-lbl lbl bb-left)))
          (insert-later! (prec-bb bb bb-left) bb-left
            (lambda (x bb-left) (queue-put! bb-queue bb) (insert! bb bb-left))))
        (insert! #f bb-left)))

    (insert-now! (find-block entry-lbl) bb-list)

    (queue->list bb-queue)))
